using System;
using System.Collections.Generic;
using System.Text;
using System.Net.Sockets;
using System.Runtime.InteropServices;

namespace novaterm
{
	class novacom
	{
		public novacom() 
		{
		}

		#region Public Methods

		public Device[] GetDeviceList()
		{
			String deviceListString;
			Device[] devices = null;

			try {
				TcpClient client = new TcpClient(_serverAddress, _deviceListPort);

				NetworkStream stream = client.GetStream();

				byte[] data = new byte[1024];

				int pos = 0;
				while (pos < data.Length) {
					int bytesRead = stream.Read(data, pos, data.Length - pos);
					if (bytesRead == 0)
						break;

					pos += bytesRead;
				}

				stream.Close();

				deviceListString = System.Text.Encoding.ASCII.GetString(data, 0, pos);
			} catch (Exception) {
				return null;
			}

			// parse the device list
			string[] lines = deviceListString.Split('\n');

			foreach (string line in lines) {
				if (line.Length > 0) {
					Console.WriteLine(line);

					string[] fields = line.Split(' ');
					
					// XXX hack that returns only one
					devices = new Device[1];
					devices[0] = new Device(Convert.ToInt32(fields[0]), (string)fields[1].Clone(), _serverAddress);
					break;
				}
			}

			return devices;
		}

		public Connection Connect(Device dev) 
		{
			Connection c = new Connection(dev);

			c.Connect();

			return c;
		}

		#endregion


		#region Properties
		public String ServerAddress 
		{
			get
			{
				return this._serverAddress;
			}
			set
			{
				_serverAddress = value;
			}

		}

		public int DeviceListPort
		{
			get
			{
				return this._deviceListPort;
			}
			set
			{
				_deviceListPort = value;
			}

		}
		#endregion

		private String _serverAddress;
		private int _deviceListPort = 6968;

		public class Device
		{
			public Device()
			{
			}

			public Device(int _port, string _nduid, string _server)
			{
				port = _port;
				nduid = _nduid;
				server = _server;
			}

			public int port;
			public String nduid;
			public String server;
		};

		public class Connection
		{
			public Connection(Device _dev)
			{
				dev = _dev;
			}

			public void Connect()
			{
				client = new TcpClient(dev.server, dev.port);
				stream = client.GetStream();
			}

			public void SendCommand(string command) 
			{
				byte[] smk = new byte[command.Length + 1];

				for (int i=0; i < command.Length ; i++)
				{
					System.Byte ss = Convert.ToByte(command[i]);
					smk[i] = ss ;
				}
				smk[command.Length] = Convert.ToByte('\n');

				stream.Write(smk, 0, smk.Length);
			}

			public int GetCommandResponse()
			{
				char[] buf = new char[1024];

				int i;
				for (i = 0; i < buf.Length; i++) {
					buf[i] = Convert.ToChar(stream.ReadByte());
					if (Convert.ToChar(buf[i]) == '\n')
						break;
				}

				string response = new String(buf, 0, i);
				string[] resp = response.Split(' ');
				if (resp[1] == "0")
					return 0;

				return -1;
			}

			public void SendData(byte[] buf)
			{
				if (packet != null)
					packet.Write(buf, 0, buf.Length);
			}

			public void SendData(string str)
			{
				byte[] smk = new byte[str.Length];

				for (int i = 0; i < str.Length; i++) {
					System.Byte ss = Convert.ToByte(str[i]);
					smk[i] = ss;
				}

				if (packet != null)
					packet.Write(smk, 0, smk.Length);
			}

			public void SetWindowSize(int cols, int rows)
			{
				if (packet != null)
					packet.SendResize(cols, rows);
			}

			private void Received(byte[] buf, int offset, int length)
			{

			}

			private void SocketAsyncReadCallback(IAsyncResult ar)
			{
				int nBytesRec = stream.EndRead(ar);

				if (nBytesRec > 0) {
/*
					byte[] buf = new byte[nBytesRec];

					Array.Copy(asyncBuf, 0, buf, 0, nBytesRec);
 */
					packet.ReadHook(asyncBuf, 0, nBytesRec);
				}
				stream.BeginRead(asyncBuf, 0, asyncBuf.Length, SocketAsyncReadCallback, this);
			}

			public void StartAsyncConnection(AsyncReadCallback callback)
			{
				if (asyncBuf == null)
					asyncBuf = new byte[4096];


				packet = new Packetizer(stream);
				packet.readCallback += callback;
				stream.BeginRead(asyncBuf, 0, asyncBuf.Length, SocketAsyncReadCallback, this);
			}

			public delegate void AsyncReadCallback(byte[] data, int offset, int size, stream_type type);

			private Device dev;
			private TcpClient client;
			private NetworkStream stream;
			private byte[] asyncBuf;
			private Packetizer packet;
		};

		private class Packetizer 
		{
			public Packetizer(NetworkStream _stream)
			{
				stream = _stream;
				headerBuffer = new byte[16];
				headerPos = 0;
				dataPos = 0;
				dataLen = 0;
			}

			private void WriteObject(object header)
			{
				int s = Marshal.SizeOf(header);
				IntPtr ip = Marshal.AllocHGlobal(s);
				Marshal.StructureToPtr(header, ip, false);

				byte[] headerBuf = new byte[s];
				Marshal.Copy(ip, headerBuf, 0, s);
				Marshal.FreeHGlobal(ip);

				stream.Write(headerBuf, 0, headerBuf.Length);
			}
			
			public void Write(byte[] buffer, int offset, int size)
			{
				packetHeader header;

				header.magic = packetHeaderMagic;
				header.version = packetHeaderVersion;
				header.size = (UInt32)size;
				header.type = (UInt32)packetType.PACKET_TYPE_DATA;

				WriteObject(header);
				stream.Write(buffer, offset, size);
			}

			public void SendResize(int cols, int rows)
			{
				packetHeader header;
				header.magic = packetHeaderMagic;
				header.version = packetHeaderVersion;
				header.size = (UInt32)oobHeaderLen;
				header.type = (UInt32)packetType.PACKET_TYPE_OOB;
				WriteObject(header);

				packetOOBMessageResize resize;
				resize.message = (UInt32)packetMessage.PACKET_OOB_RESIZE;
				resize.cols = (UInt32)cols;
				resize.rows = (UInt32)rows;
				resize.pad = 0;
				resize.pad2 = 0;
				WriteObject(resize);
			}

			public void ReadHook(byte[] buffer, int offset, int size)
			{
				while (offset < size) {
					// see if we have a complete header yet
					if (headerPos < headerLen) {
						int i;
						int tocopy;

						tocopy = headerLen - headerPos;
						if (tocopy > (size - offset))
							tocopy = size - offset;

						for (i = headerPos; i < tocopy; i++) {
							headerBuffer[headerPos] = buffer[offset];
							headerPos++;
							offset++;
						}

						// we just completed a header
						if (headerPos == headerLen) {
							IntPtr ip = Marshal.AllocHGlobal(headerLen);
							Marshal.Copy(headerBuffer, 0, ip, headerLen);

							packetHeader header = (packetHeader)Marshal.PtrToStructure(ip, typeof(packetHeader));
							Marshal.FreeHGlobal(ip);

							// XXX validate header
							dataPos = 0;
							dataLen = (int)header.size;
						}
					}

					if (dataPos < dataLen) {
						int toread;

						toread = dataLen - dataPos;
						if (toread > (size - offset))
							toread = size - offset;

						if (readCallback != null) {
							readCallback(buffer, offset, toread, stream_type.STREAM_STDOUT);
						}
						offset += toread;
						dataPos += toread;

						// see if we're done with data and need to reenter header phase
						if (dataPos == dataLen) {
							dataLen = 0;
							headerPos = 0;
						}
					}
				}
			}


			private NetworkStream stream;
			public event Connection.AsyncReadCallback readCallback;

			private int dataPos;
			private int dataLen;
			private int headerPos;
			private byte[] headerBuffer;

			/* on the wire packet header */
			struct packetHeader {
				public UInt32 magic;
				public UInt32 version;
				public UInt32 size;
				public UInt32 type;	
			};

			const UInt32 packetHeaderMagic = 0xdecafbad;
			const UInt32 packetHeaderVersion = 1;
			const int headerLen = 16;

			enum packetType
			{
				PACKET_TYPE_DATA = 0,
				PACKET_TYPE_ERR = 1,
				PACKET_TYPE_OOB = 2,
			};

			enum packetMessage
			{
				PACKET_OOB_EOF = 0,
				PACKET_OOB_SIGNAL = 1,
				PACKET_OOB_RETURN = 2,
				PACKET_OOB_RESIZE = 3,
			};

			struct packetOOBMessageResize {
				public UInt32 message; // PACKET_OOB_RESIZE

				public UInt32 rows;
				public UInt32 cols;
				public UInt32 pad;
				public UInt32 pad2;
			};

			const int oobHeaderLen = 20;
		};

		public enum stream_type {
			STREAM_STDIN,
			STREAM_STDOUT,
			STREAM_STDERR,
		}
	}

}
